# fmt: off
__version__ = '12.0.278'
"""## Now \\*should\\* not behave like shit with pyinstaller, but idk really.

This it not guaranteed, but strongly believed by me to always be a str with valid version info...
"""
