# fmt: off
BACKSLASH  = '\\'
NEWLINE    = '\n'
CARR_RET   = '\r'
BACKSPACE  = '\b'
BACKTICK   = '`'
BACKTICKS  = 3*BACKTICK
APOSTROPHE = "'"
QUOTE      = '"'
FAKE_PIPE  = '¦'

# fmt: off
__all__ = [
  "BACKSLASH",
  "NEWLINE",
  "CARR_RET",
  "BACKSPACE",
  "BACKTICK", "BACKTICKS",
  "APOSTROPHE", "QUOTE",
  "FAKE_PIPE",
]
